;===================================================
; This is the MASM 6+ source code for a QuickBASIC
; Mouse procedure library. Beyond providing wrappers
; for a few of the most common mouse functions, it
; includes a MouseInit procedure that initializes
; the mouse driver and installs a mouse driver
; interrupt subroutine that maintains the current
; mouse status in a QuickBASIC global variable.
;
; The QuickBASIC global variable must be of type
; MouseType, and the segment and offset addresses
; of the variable must be passed in the first call
; to the MouseInit procedure. The button states
; are TRUE (-1) if the button is pressed, or FALSE
; (0) if the button is not pressed. The event mask
; indicates what type of mouse event triggered the
; most recent interrupt. The possible events are a
; position change or a press or release of the
; left or right button.
;
; TYPE MouseType
;  x        AS INTEGER ' Cursor X coordinate
;  y        AS INTEGER ' Cursor Y coordinate
;  left     AS INTEGER ' Left button state
;  right    AS INTEGER ' Right button state
;  event    AS INTEGER ' Event mask
; END TYPE
; DIM SHARED mouse AS MouseType
;
; These constants are used to interpret the event
; mask. The assigned values are the value of the
; corresponding bit in the event mask, so an
; event can be detected by ANDing the appropriate
; constant with the event mask.
; CONST POSITION = 1
; CONST LPRESS = 2
; CONST LRELEASE = 4
; CONST RPRESS = 8
; CONST RRELEASE = 16
;
; For the range of video modes that are supported
; by QuickBASIC, the interrupt subroutine and the
; MouseSetPosition procedure automatically
; translate between mouse driver virtual screen
; coordinates and physical screen coordinates,
; so as viewed from the QuickBASIC module all
; coordinates are physical screen coordinates.
; For the text modes, the x coordinate will be
; the base 1 column position and the y coordinate
; will be the base 1 row position.
;
; For uniformity, all of the procedures that take
; arguments expect the arguments to be passed by
; reference. "Pass by reference" is the QB default
; and it means that the value passed is the address
; of the argument in the QB default data segment.
; Within the procedures, the arguments, which as
; viewed from the procedure are properly termed
; "parameters", are accessed by first loading the
; address of the parameter into a base or index
; register and then by using a register indirect
; form of the instruction that accesses the
; parameter. For example, to load the value of a
; parameter named varPtr into AX:
;
;     mov   bx,varPtr
;     mov   ax,[bx]
;
; QB procedures are always called with a far call
; and the procedures must preserve the direction
; flag and the BP, DI, SI, DS, and SS registers.
; The BASIC calling conventions require that the
; arguments be pushed onto the stack in left to
; right order as they appear in the procedure
; definition, and that the called procedure remove
; the arguments from the stack. One significant
; advantage of using MASM for mixed language
; programming is that you can specify a language
; type in the .MODEL directive and MASM will
; automatically generate the code that is required
; to properly implement the calling conventions.
; For example, for the RET instruction in the
; procedures MASM knows to encode a far return,
; and for the procedures that take arguments, MASM
; knows to add an operand to the RET instruction
; that will cause the processor to add an
; appropriate value to SP after the procedure has
; returned to the caller.
;===================================================

; Declare a structure to use as a template when
; accessing the mouse status variable.

MouseType struct
   x        WORD ?
   y        WORD ?
   left     WORD ?
   right    WORD ?
   event    WORD ?
MouseType ends

; This prototype establishes the call interface for
; the B_OnExit procedure (so MASM will know how to
; call it) and effectively generates an external
; declaration in this module (so MASM will know that
; the procedure is defined in another module).

B_OnExit proto far basic :far ptr

; This is the normal model specification for QB.

.model medium,basic

; Enable assembly of the 186 instruction set (the
; minimum processor that will support an immediate
; (constant) operand for a push instruction).

.186

; Start a near data segment. The linker will combine
; this segment with the QB default data segment.

.data
   ; Allocate a flag variable to lock out multiple
   ; calls to MouseInit. The name avoids the MASM
   ; reserved word "finit".

   f_init   dw 0

   ; Define screen and cursor masks for a crosshair
   ; cursor. The cursor defintion is within this
   ; module rather than being passed as a parameter
   ; simply because defining the cursor with binary
   ; numbers is much easier than defining it with
   ; hex numbers.
   ;
   ; Note that this cursor will not work correctly
   ; for SCREEN 1, and that the aspect ratio will
   ; be significantly off for the 640x200 modes. 
   ;
   ; For the QB graphics modes other than SCREEN 1,
   ; the mouse driver will AND the screen mask bits
   ; with the corresponding screen pixel bits and
   ; XOR the result with the cursor mask. In truth
   ; table form:
   ;
   ;  screen mask cursor mask resulting screen bit
   ;       0            0              0
   ;       0            1              1
   ;       1            0          unchanged
   ;       1            1          inverted
   ;
   ; Note that the masks bits are expanded as
   ; necessary for the current graphics mode.
   ; For example, for mode 13h (SCREEN 13) each
   ; mask bit is expanded to 8 bits and these
   ; bits are then combined with the 8 attribute
   ; bits for the corresponding screen pixel.

              ;0123456701234567
   cross    dw 1111111111111111b ;0
            dw 1111111111111111b ;1
            dw 1111111011111111b ;2
            dw 1111111011111111b ;3
            dw 1111111011111111b ;4
            dw 1111111011111111b ;5
            dw 1111111111111111b ;6
            dw 0000001110000001b ;7
            dw 1111111111111111b ;0
            dw 1111111011111111b ;1
            dw 1111111011111111b ;2
            dw 1111111011111111b ;3
            dw 1111111011111111b ;4
            dw 1111111111111111b ;5
            dw 1111111111111111b ;6
            dw 1111111111111111b ;7

            dw 0000000000000000b ;0
            dw 0000000000000000b ;1
            dw 0000000100000000b ;2
            dw 0000000100000000b ;3
            dw 0000000100000000b ;4
            dw 0000000100000000b ;5
            dw 0000000000000000b ;6
            dw 1111110001111110b ;7
            dw 0000000000000000b ;0
            dw 0000000100000000b ;1
            dw 0000000100000000b ;2
            dw 0000000100000000b ;3
            dw 0000000100000000b ;4
            dw 0000000000000000b ;5
            dw 0000000000000000b ;6
            dw 0000000000000000b ;7

; Start a (far) code segment.

.code

; This include directive include TEST.ASM in this
; file, effectively making it part of this file.

include test.asm

;===================================================

; Allocate a variable to store the far address of
; the mouse status variable, and a lookup table of
; shift and base adjust values indexed by BIOS video
; mode, that will be used to translate between mouse
; driver virtual screen coordinates and physical
; screen coordinates. These variables need to be in
; the code segment because they must be accessed
; from the interrupt subroutine, and when the mouse
; driver calls the subroutine the only segment
; register with a known value is CS.

statusVarPtr   dd 0

; To allow the table to be indexed by the BIOS video
; mode the table must include all mode numbers from
; 0 to 13h, even though several of the included
; modes have no corresponding QB SCREEN mode and
; others are not supported on the typical PC. Each
; element consists of an x shift word, a y shift
; word, a base adjust word, and a pad word that is
; included to simplify table indexing. The base
; adjust value is 1 for the text modes and 0 for
; the graphic modes. Aligning the table on a word
; boundary minimizes the time required to access
; the table.

align 2
luTable  dw 4,3,1,0  ; 0
         dw 4,3,1,0  ; 1 (SCREEN 0, 40 column)
         dw 3,3,1,0  ; 2
         dw 3,3,1,0  ; 3 (SCREEN 0, 80 column)
         dw 1,0,0,0  ; 4 (SCREEN 1)
         dw 1,0,0,0  ; 5
         dw 0,0,0,0  ; 6 (SCREEN 2)
         dw 3,3,1,0  ; 7 (SCREEN 0, monochrome only)
         dw 2,0,0,0  ; 8 PCjr only
         dw 0,0,0,0  ; 9 PCjr only
         dw 0,0,0,0  ; A PCjr only
         dw 0,0,0,0  ; B EGA BIOS internal only
         dw 0,0,0,0  ; C EGA BIOS internal only
         dw 1,0,0,0  ; D (SCREEN 7)
         dw 0,0,0,0  ; E (SCREEN 8)  
         dw 0,0,0,0  ; F (SCREEN 10) 
         dw 0,0,0,0  ; 10 (SCREEN 9)
         dw 0,0,0,0  ; 11 (SCREEN 11)
         dw 0,0,0,0  ; 12 (SCREEN 12)
         dw 1,0,0,0  ; 13 (SCREEN 13)
;===================================================

; The following procedure declarations use the
; distance (far) and langtype (basic) from the
; .MODEL directive, and default to PUBLIC
; visibility.

;===================================================
; This proc calls the mouse driver Mouse Reset and
; Status function to reset the mouse driver to clear
; the interrupt subroutine call mask, which causes
; the mouse driver to cease calling the interrupt
; subroutine.

; This declaration must be placed before the
; reference to it in the MouseInit procedure.
;===================================================
TermProc proc
      xor   ax,ax
      int   33h
      ret
TermProc endp

;===================================================
; This proc checks for a mouse driver, resets it,
; saves the segment and offset addresses of the
; global status variable, installs the interrupt
; subroutine, and calls the B_OnExit routine to log
; a termination procedure that will automatically
; disable the interrupt subroutine when the program
; terminates. It returns non-zero for success, or
; zero for failure. The first call sets a flag that
; locks out subsequent calls.
;===================================================
MouseInit proc varSeg:WORD,varPtr:WORD

      ; Lock out subseqent calls.
      .IF f_init != 0
         xor   ax,ax
         ret
      .ENDIF
      mov   f_init,-1

      ; Check for a mouse driver.

      ; Use the DOS Get Interrupt Vector function
      ; to get the interrupt 33h vector. If the
      ; segment address is zero, then the mouse
      ; driver is not installed.
      mov   ax,3533h
      int   21h
      mov   ax,es
      .IF ax == 0
         ret
      .ENDIF

      ; Attempt to reset the mouse driver. If the
      ; reset fails then the mouse driver is not
      ; installed.
      xor   ax,ax
      int   33h
      .IF ax == 0
         ret
      .ENDIF

      ; Save the address of the mouse status variable.
      mov   bx,varPtr
      mov   ax,[bx]
      mov   word ptr cs:statusVarPtr,ax
      mov   bx,varSeg
      mov   ax,[bx]
      mov   word ptr cs:statusVarPtr+2,ax

      ; Install the interrupt subroutine.
      mov   ax,12
      ; This condition mask specifies that an
      ; interrupt be generated for any of the
      ; defined conditions.
      mov   cx,11111b
      push  cs
      pop   es
      mov   dx,OFFSET InterruptSub
      int   33h

      ; Register the termination routine.

      ; The arguments must be pushed onto the stack
      ; as per the basic calling convention before
      ; calling the routine.
      push  cs
      push  OFFSET TermProc
      call  B_OnExit

      ; Return whatever B_OnExit returned.
      ret
MouseInit endp

;===================================================
; This proc is the interrupt subroutine. The mouse
; driver will call it for each mouse interrupt.
;
; The local variables allow temporary values to be
; stored on the stack in named locations.
;===================================================
align 2
InterruptSub proc

      LOCAL xShift:WORD,yShift:WORD,baseAdjust:WORD

      ; Preserve DS before changing it because the
      ; mouse driver probably will not expect it to
      ; change.
      push  ds

      ; Preserve BX before changing it because it
      ; contains the button status.
      push  bx

      ; Get the video mode from the BIOS data area.
      mov   bx,40h
      mov   ds,bx
      mov   bx,49h
      mov   bx,[bx]
      ; Because the mode is actually stored as a
      ; byte and we are reading a word, and because
      ; bit 7 of the byte may or may not be set
      ; depending on whether or not the display
      ; memory was erased during the most recent
      ; mode set, we need to discard the upper 9
      ; bits of BX.
      and   bx,1111111b

      ; Get the corresponding x and y shift and
      ; base adjust values and store them in
      ; local variables. BX must first be scaled by
      ; the size of the table elements (BX=BX*8).
      shl   bx,3
      ; For instructions that take two operands,
      ; only one can be a memory operand. The push-
      ; pop sequences are used to perform a memory
      ; to memory move without involving a scratch
      ; register. Local variables can be accessed
      ; without a segment override because they are
      ; allocated from the stack, so the POP
      ; instructions reference BP (for example,
      ; "pop xShift" is encoded as "pop [bp-2]"),
      ; so the processor automatically uses SS.
      push  cs:[luTable+bx]
      pop   xShift
      push  cs:[luTable+bx+2]
      pop   yShift
      push  cs:[luTable+bx+4]
      pop   baseAdjust

      ; Load the status variable address into DS:BX.      
      mov   bx,word ptr cs:[statusVarPtr]
      mov   ds,word ptr cs:[statusVarPtr+2]

      ; This assume informs MASM that BX contains
      ; a pointer to a variable of MouseType, so
      ; it will know what the following references
      ; to [bx].elementname mean:
      ASSUME bx:ptr MouseType

      ; Store the event mask in the status variable.
      mov   [bx].event,ax

      ; We are Finished with the value in AX so pop
      ; the preserved button status into it.
      pop   ax

      ; Store the button status in the status
      ; variable. Bit0 of the button status will
      ; be set if the left mouse button is pressed
      ; and bit1 if the right mouse button is
      ; pressed. In the MASM 6+ High Level syntax
      ; "&" is the bit test operator, which returns
      ; true if the bit with the specified "place"
      ; value is set.
      .IF ax & 1
         mov   [bx].left,-1
      .ELSE
         mov   [bx].left,0
      .ENDIF
      .IF ax & 2
         mov   [bx].right,-1
      .ELSE
         mov   [bx].right,0
      .ENDIF

      ; Convert the virtual screen coordinates
      ; of the mouse cursor to physical screen
      ; coordinates and store the results in
      ; the status variable. 
      mov   ax,cx
      mov   cx,xShift
      shr   ax,cl
      add   ax,baseAdjust
      mov   [bx].x,ax

      mov   ax,dx
      mov   cx,yShift
      shr   ax,cl
      add   ax,baseAdjust
      mov   [bx].y,ax

      ; Remove the assumption for BX.
      ASSUME bx:NOTHING

      ; Recover DS.
      pop   ds

      ret
InterruptSub endp

;===================================================
; This proc calls the mouse driver Show Cursor
; function. This function causes the mouse driver
; to increment its internal cursor flag and, if
; the flag is 0 after being incremented, display
; the mouse cursor. The flag will never be greater
; than 0.
;===================================================
MouseShow proc
      mov   ax,1
      int   33h
      ret
MouseShow endp

;===================================================
; This proc calls the mouse driver Hide Cursor
; function. This function causes the mouse driver
; to remove the mouse cursor from the screen and
; decrement its internal cursor flag.
;
; Note that the mouse driver decrements the flag
; with each call even if the mouse cursor is hidden.
; If this procedure is called multiple times without
; intervening calls to ShowMouseCursor, then the
; mouse driver will not redisplay the cursor until
; ShowMouseCursor has been called an equal number
; of times.
;===================================================
MouseHide proc
      mov   ax,2
      int   33h
      ret
MouseHide endp

;===================================================
; This proc calls the mouse driver Set Mouse Cursor
; Position function to set the mouse cursor position
; to the specified physical screen coordinates.
;
; Refer to the InterruptSub for an explanation of
; the mouse driver virtual screen coordinate to
; physical screen coordinate conversion, reversed
; in this procedure.
;===================================================
MouseSetPosition proc x:WORD,y:WORD

      LOCAL xShift:WORD,yShift:WORD,baseAdjust:WORD
      
      mov   bx,40h
      mov   es,bx
      mov   bx,49h
      mov   bx,es:[bx]    
      and   bx,1111111b
      shl   bx,3

      push  cs:[luTable+bx]
      pop   xShift
      push  cs:[luTable+bx+2]
      pop   yShift
      push  cs:[luTable+bx+4]
      pop   baseAdjust
      
      mov   bx,y
      mov   ax,[bx]
      mov   cx,yShift
      shl   ax,cl
      sub   ax,baseAdjust
      mov   dx,ax

      mov   bx,x
      mov   ax,[bx]
      mov   cx,xShift
      shl   ax,cl
      sub   ax,baseAdjust
      mov   cx,ax

      mov   ax,4
      int   33h
      ret
MouseSetPosition endp

;===================================================
; This proc calls the mouse driver Set Graphics
; Cursor Block function to define the shape and
; center of the graphics cursor.
;===================================================
MouseSetCrossCursor proc
      mov   ax,9
      ; BX and CX specify the cursor hot spot.
      mov   bx,7
      mov   cx,7
      push  ds
      pop   es
      mov   dx,OFFSET cross
      int   33h
      ret
MouseSetCrossCursor endp

end
