;===================================================
; This file contains a procedure that performs a
; measurement of the number of clock cycles the
; interrupt subroutine takes to execute. To use the
; procedure, include this file in MOUSELIB.ASM
;
; The purpose of the measurement was to ensure that,
; in the worst case, the interrupt subroutine would
; not consume an unreasonable amount of processor
; time. As it is currently coded, and running on a
; P3, the interrupt subroutine takes ~120 clocks
; per call. In my tests, the mouse driver generated
; interrupts at a maximum rate of 100 per second.
; Assuming a 100MHz processor clock and the maximum
; mouse driver interrupt rate, the interrupt
; subroutine would consume 120 micro seconds per
; second, or 0.012% of the total processor time.
;
; The number of clock cycles is returned in AX, so
; this proc should be declared as a function:
;
;  DECLARE FUNCTION InterruptSubClocks
;
; This procedure requires a Pentium class processor,
; and calling it on a system that does not meet this
; requirement will probably result in some form of
; undesirable behavior.
;
; The rdtsc (Read Time Stamp Counter) instruction
; returns the current value of the 64-bit Time Stamp
; Counter in EDX:EAX. The Time Stamp Counter resets
; to zero when the processor is reset (a reset also
; occurs at power up) and increments on every clock
; cycle.
;
; The cpuid instruction is used here to prevent the
; rdtsc instruction from executing in parallel with
; other instructions.
;===================================================
.586
InterruptSubClocks proc

      LOCAL startCnt:QWORD,overhead:QWORD

      ; Return zero if statusVarPtr is zero,
      ; indicating that the MouseInit procedure
      ; was not called or was not successful.
      ; If the interrupt subroutine were called
      ; with statusVarPtr set to zero, it would
      ; overwrite the bottom 10 bytes in the real
      ; mode interrupt vector area, which would
      ; cause a crash the first time interrupt 0,
      ; 1 or 2 was called.
      cmp   cs:statusVarPtr,0
      jne   @F
      mov   ax,0
      ret
   @@:

      ; Determine the overhead clocks by timing
      ; an empty loop.
      xor   eax,eax
      cpuid
      rdtsc
      mov   dword ptr [startCnt],eax
      mov   dword ptr [StartCnt+4],edx
      xor   eax,eax
      cpuid

      mov   cx,1000
   @@:
      push  cx
      ; No call here.
      pop   cx
      sub   cx,1      
      jnz   @B

      xor   eax,eax
      cpuid
      rdtsc
      sub   eax,dword ptr [startCnt]
      sbb   edx,dword ptr [startCnt+4]
      mov   dword ptr [overhead],eax
      mov   dword ptr [overhead+4],edx

      ; Get to total clocks for the loop, this
      ; time with the call to InterruptSub.
      xor   eax,eax
      cpuid
      rdtsc
      mov   dword ptr [startCnt],eax
      mov   dword ptr [StartCnt+4],edx
      xor   eax,eax
      cpuid
      mov   cx,1000
   @@:
      push  cx
      call InterruptSub
      pop   cx
      sub   cx,1
      jnz   @B
      xor   eax,eax
      cpuid
      rdtsc
      sub   eax,dword ptr [startCnt]
      sbb   edx,dword ptr [startCnt+4]

      ; Subtract the overhead.
      sub   eax,dword ptr [overhead]
      sbb   edx,dword ptr [overhead+4]

      ; Divide EDX:EAX by the number of loops.
      ; The quotient in EAX will be well within
      ; the valid range of a QB integer, which
      ; QB expects in AX.
      mov   ebx,1000
      div   ebx

      ret
InterruptSubClocks endp
;===================================================
